// Stereoptica V2.2

/* ********************************************************* Start legal stuff

- Copyright (c) 2007+, Daniel White, Skytopia.com
- All rights reserved.
- 
- --- Results (rendered pictures) produced from the program (the binary/exe)
- --- may be freely used providing:
- 1: No more than five frames (pictures) are used for commercial purposes,
- without my express permission. Anyone wishing to do more than five in total
- may wish to do so for a reasonable fee by contacting me from:
- http://www.skytopia.com/email.html
-
- Furthermore, you may modify the source for your own use (the above clause
- still applies), but redistribution of the modified source is not allowed
- without my permission. However, you may use up to 50 lines of code (except
- the exact code in the "POST PROCESSING" section), or any of the function
- code in your own projects freely.
- 
- --- Redistribution of the unmodified program is permitted, providing:
- 1: The software is not sold or used for commercial purposes, without my
- express permission
- 2: All of the files are kept together and intact.
-
- THIS SOFTWARE IS PROVIDED BY DANIEL WHITE "AS IS" AND ANY
- EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- DISCLAIMED. IN NO EVENT SHALL DANIEL WHITE BE LIABLE FOR ANY DIRECT,
- INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
- BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
- OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
- EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

****************************

This software uses the (excellent) FreeImage open source image library.
See http://freeimage.sourceforge.net for details.

FreeImage is used under the (GNU GPL or FIPL), version 1.0.

Documents available at:
http://ignum.dl.sourceforge.net/project/freeimage/Source%20Documentation/3.13.1/FreeImage3131.pdf

****************************

Through FreeImage, this software also uses the amazing NeuQuant neural-net
quantization algorithm by Anthony Dekker for converting 24 bit images to 8 bit.

********************************************************* End of legal stuff*/

/////////////////////// To Compile:
// Add  -lFreeImage to CFLAGS in makefile
// Copy FreeImage.lib to Dev-Cpp lib folder
// Copy FreeImage.h to Stereoptica folder
//
// Just FreeImage.dll is needed for final exe


// *nix users. You'll need to download the source from
// http://freeimage.sourceforge.net/ to be able to compile.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "FreeImage/FreeImage.h"

struct RGB { double r; double g; double b; };
struct HSV { double h; double s; double v; };

int mini(int a,int b) {
    if (a<=b) return a;
    else return b;
}

char* incrementFile(int i, char*d);
char* getFileNumber(int i, char*d);

int countFiles(char*filename) {
	FILE *fs;
	int count=0;
	fs = fopen(filename, "rb");

	
	while (  fs ) {


		count++;
		//fclose(fs);
		filename = incrementFile(1,filename);
		fs = fopen(filename, "rb");
	}


	return count;
}

RGB toRGB(HSV z);
HSV toHSV(RGB z);

double mod(double a, double b);
char* arrayc(int n)  { return (char*)malloc(n*sizeof(char));}
char* int2string(int n) {
	char*s = arrayc(20);
	sprintf(s, "%d", n);
	return s;
}
void writefile(char* text, char*filename) {
    int len = strlen(text);
	
	FILE *fs = fopen(filename, "wb");
	if (!fs) { printf("Can't write file"); exit(0); }
	fputs(text,fs);
	fclose(fs);	
}


int main(int argc, char *argv[])
{


	
	


	// Global settings:
	bool stereo=0;	// Defaults to 0 - anaglyph mode (3D glasses). Select 1 for stereo.
	bool batch=0;		// 0 is default, since it'll look for any extra frames.
	int trimOut=0;		// Usual one to trim. 10 or 20 (pixels) is often a good choice. 0 = default.
	int trimIn=0;
	char left[500];
	char right[500];
	char saveFormat[3];	// png, gif, bmp, jpg
	int col256=0;
	bool swapxy=0; // Default=0
	int frames=0;
	int actualFrames;

	// Anaglyph settings (only applicable if stereo=0):
//	double cyanChange=1.0; // Typical values range from 0.5 to 1. Default = 1 // *** DISCONTINUED ***
	double redChange=1.4;  // Typical values range from 1 to 3. Default = 1.4
	bool mono=0;
	bool noEnhance=0; // 1 turns off experimental variable which allows less 'ghosting'.
	double evar1=1.2; // For use if noEnhance remains at 0. 1.2 is default.
	double evar2=0.2; // For use if noEnhance remains at 0. 0.2 is default.

	// Stereo settings (only applicable if stereo=1):
	int centerGap=0; // Default=0

	// Stereo settings - Center Gap Colour settings (only applicable if stereo=1 & centerGap>0):
	int gapr=0; // Red element: min 0, max 255
	int gapg=0; // Green element: min 0, max 255
	int gapb=0; // Blue element: min 0, max 255

	// Pre processing of images:
	double hueChange=0; // 0 to 6 cycles through hue colour wheel.
	double brightChange=1.0; // Change brightness/value (1=no change)
	double chromaChange=1.0; // Change chroma/saturation (1=no change)
	bool reverseHues=0; // RGB becomes BGR. Use in conjunction with hueChange to cycle colours.

	bool returnDimensions=0;

	// Initialize some default values;
	left[0]='\0';
	right[0]='\0';
	strcpy(saveFormat,"png");

	// Copy command line parameter to variables if they exist
	for (int n=1; n<argc; n++) {

		//printf("arg %i: %s_\n",n,argv[n]);
		

		char* token;
		char var[50];
		token = strtok (argv[n],"=");
		strcpy(var,token);
		token = strtok (NULL,"=");

		//printf("%i var: %s\n",n,var);
		//printf("%i token: %s\n",n,token);


		
		if (strcmp(var,"stereo")==0 ) stereo=1;
		if (strcmp(var,"trimout")==0  ) trimOut=atoi(token);
		if (strcmp(var,"trimin")==0  ) trimIn=atoi(token);

		if (strcmp(var,"l")==0  ) strcpy(left,token);
		if (strcmp(var,"r")==0  ) strcpy(right,token);


		if (strcmp(var,"countf")==0) {
			int countLeftFrames =  countFiles(left);
			int countRightFrames = countFiles(right);
			int min = mini(countLeftFrames, countRightFrames);
			if (countLeftFrames!=countRightFrames) min=-min;
			
			writefile(int2string(min),"info.txt");

			exit(0);
		}



		if (strcmp(var,"save")==0  ) strcpy(saveFormat,token);
		if (strcmp(var,"frames")==0  ) frames = atoi(token);

		if (strcmp(var,"batch")==0) batch=1;




		if (strcmp(var,"col256")==0  ) col256 = 1;

		if (strcmp(var,"monochrome")==0  ) mono = 1;
		if (strcmp(var,"red")==0  ) redChange = atof(token);
//		if (strcmp(var,"cyanchange")==0  ) cyanChange = atof(token);
		if (strcmp(var,"noenhance")==0  ) noEnhance = 1;
		if (strcmp(var,"evar1")==0  ) evar1 = atof(token);
		if (strcmp(var,"evar2")==0  ) evar2 = atof(token);

		if (strcmp(var,"gap")==0  ) centerGap = atoi(token);
		if (strcmp(var,"swap")==0  ) swapxy = 1;



		if (strcmp(var,"gapr")==0  ) gapr = atoi(token);
		if (strcmp(var,"gapg")==0  ) gapg = atoi(token);
		if (strcmp(var,"gapb")==0  ) gapb = atoi(token);

		if (strcmp(var,"hue")==0  ) hueChange = atof(token);
		if (strcmp(var,"bright")==0  ) brightChange = atof(token);
		if (strcmp(var,"chroma")==0  ) chromaChange = atof(token);
		if (strcmp(var,"reversehues")==0  ) reverseHues = 1;

		if (strcmp(var,"help")==0 || strcmp(var,"-help")==0 || strcmp(var,"-h")==0 || strcmp(var,"h")==0 ) {
			printf("Typical use: stereoptica l=myleftpic.png r=myrightpic.png\n\n");
			printf("Another example: stereoptica l=myleftpic.jpg r=myrightpic.jpg stereo batch trimout=20 bright=1.5\n\n");
			printf("Options: (replace *** stars with number or name)\n");

			printf("\n  Global settings:");
			printf("\nl=***          Replace *** with filename of left pic. E.g.: l=myleftpic.gif");
			printf("\nr=***          Replace *** with filename of your right picture");
			printf("\nsave=***       Enter gif, png, jpg, or bmp for save format (default = png)");
			printf("\ncol256         Convert from true colour to 256 colours before saving");
			printf("\nstereo         Add this parameter for a stereogram instead of an anaglyph");
			printf("\nswap           Swaps the left and right pictures (easy fix for bad looking 3D)");
			printf("\ntrimout=***    Measure in pixels. Crops the outside of pictures (default = 0)");
			printf("\ntrimin=***     Measure in pixels. Try trimout before this (default = 0)");
			printf("\nbatch          Batch process multiple images. Read help doc for more info");
			printf("\nframes=***     No. of frames to process if batch is selected (default=0 (all))\n");
			
			printf("\n  Further global settings - Pre-processing of images:");
			printf("\nhue=***        Enter from 0 to 6. Cycle hues round colour wheel (default = 0)");
			printf("\nchroma=***     Enter from 0 onwards. Adjust chroma of images (default = 1)");
			printf("\nbright=***     Enter from 0 onwards. Adjust brightness of images (default = 1)");
			printf("\nreversehues    This addition reverses the hues, so RGB becomes BGR\n");
			
			printf("\n  Anaglyph settings (ie. if 'stereo' isn't added as a variable):");
			printf("\nmonochrome     Removes colour from the anaglyph, but may have slightly nicer 3D")   ;
			printf("\nred=***        Adjust red element. Typical values range from 1-3 (default = 1.4)");
//			printf("\ncyanchange=*** Adjust cyan element. Typicallly from 0.5-1 (default = 1)");
			printf("\nnoenhance      Disable the 3D enhancing colour processing to remove ghosting");
			printf("\nevar1=***      Experimental variable 1 for 'enhance' processing (default = 1.2)");
			printf("\nevar2=***      Experimental variable 2 for 'enhance' processing (default = 0.2)\n");
			
			printf("\n  Stereo settings (ie. if 'stereo' is added as a variable):");
			printf("\ngap=***        The gap (in pixels) between the left and right images");
			printf("\ngapr=***       Set the R gap colour from 0-255 (default=0)");
			printf("\ngapg=***       Set the G gap colour from 0-255 (default=0)");
			printf("\ngapb=***       Set the B gap colour from 0-255 (default=0)\n");
			
			
			

						
			
			
			exit(0);
		}

	}

	if (batch==1) {
		if (frames==0) actualFrames = mini(  countFiles(left), countFiles(right)   ) ;
		else actualFrames = mini(   mini(  countFiles(left), countFiles(right)   ),  frames  );
	}
	else actualFrames=1;


	if (strlen(left)==0 || strlen(right)==0) {
		printf("\nMust give left and right file names as minimum. Example:\nstereoptica l=mypicl.jpg r=mypicr.jpg\nType: \"stereoptica help\"...or read the help.txt file to see all the parameters you can use.\n");
		return 0;
	}


	int frame=0;	// For batching
	int R=2;	int G=1;	int B=0;
	FIBITMAP *lbitmap; 
	FIBITMAP *rbitmap;
	FIBITMAP *sbitmap;
	FIBITMAP *wmbitmap = FreeImage_Load(FIF_PNG, "winfo", PNG_DEFAULT);
	FIBITMAP *wmsbitmap;

	char lpicformat[4];
	char rpicformat[4];

	
	lpicformat[0]=left[strlen(left)-3]; rpicformat[0]=right[strlen(right)-3];
	lpicformat[1]=left[strlen(left)-2]; rpicformat[1]=right[strlen(right)-2];
	lpicformat[2]=left[strlen(left)-1]; rpicformat[2]=right[strlen(right)-1];
	lpicformat[3]='\0'; rpicformat[3]='\0';

	// Load images:
	if (swapxy==false) {

		if ( stricmp(lpicformat,"png")==0 )			lbitmap = FreeImage_Load(FIF_PNG, left, PNG_DEFAULT);
		else if ( stricmp(lpicformat,"bmp")==0 )	lbitmap = FreeImage_Load(FIF_BMP, left, BMP_DEFAULT);
		else if ( stricmp(lpicformat,"gif")==0 )	lbitmap = FreeImage_Load(FIF_GIF, left, GIF_DEFAULT);
		else if ( stricmp(lpicformat,"peg")==0 || stricmp(lpicformat,"jpg")==0) lbitmap = FreeImage_Load(FIF_JPEG, left, JPEG_DEFAULT);
		else {writefile("load input error","info.txt"); exit(0); }

		
		if ( stricmp(rpicformat,"png")==0 )			rbitmap = FreeImage_Load(FIF_PNG, right, PNG_DEFAULT);
		else if ( stricmp(rpicformat,"bmp")==0 ) 	rbitmap = FreeImage_Load(FIF_BMP, right, BMP_DEFAULT);
		else if ( stricmp(rpicformat,"gif")==0 )		rbitmap = FreeImage_Load(FIF_GIF, right, GIF_DEFAULT);
		else if ( stricmp(rpicformat,"peg")==0 || stricmp(rpicformat,"jpg")==0) rbitmap = FreeImage_Load(FIF_JPEG, right, JPEG_DEFAULT);
		else {writefile("load input error","info.txt"); exit(0); }


	} else {


		if ( stricmp(lpicformat,"png")==0 )			lbitmap = FreeImage_Load(FIF_PNG, right, PNG_DEFAULT);
		else if ( stricmp(lpicformat,"bmp")==0 )	lbitmap = FreeImage_Load(FIF_BMP, right, BMP_DEFAULT);
		else if ( stricmp(lpicformat,"gif")==0 )	lbitmap = FreeImage_Load(FIF_GIF, right, GIF_DEFAULT);
		else if ( stricmp(lpicformat,"peg")==0 || stricmp(lpicformat,"jpg")==0) lbitmap = FreeImage_Load(FIF_JPEG, right, JPEG_DEFAULT);
		else {writefile("load input error","info.txt"); exit(0);}

		
		if ( stricmp(rpicformat,"png")==0 )			rbitmap = FreeImage_Load(FIF_PNG, left, PNG_DEFAULT);
		else if ( stricmp(rpicformat,"bmp")==0 ) 	rbitmap = FreeImage_Load(FIF_BMP, left, BMP_DEFAULT);
		else if ( stricmp(rpicformat,"gif")==0 )		rbitmap = FreeImage_Load(FIF_GIF, left, GIF_DEFAULT);
		else if ( stricmp(rpicformat,"peg")==0 || stricmp(rpicformat,"jpg")==0) rbitmap = FreeImage_Load(FIF_JPEG, left, JPEG_DEFAULT);
		else {writefile("load input error","info.txt"); exit(0);}

	
	}
	if (lbitmap==0 || rbitmap==0) {printf("\nFile names are mistyped?\n"); exit(0); }

	

	// Convert to 24 bit for processing if need be:
	printf(" %i bit color. ",FreeImage_GetBPP(lbitmap));
	if (FreeImage_GetBPP(lbitmap)!=24 || FreeImage_GetBPP(rbitmap)!=24) {
		printf(" Converting to 24 bit. ");
		lbitmap = FreeImage_ConvertTo24Bits(lbitmap);
		rbitmap = FreeImage_ConvertTo24Bits(rbitmap);
	}

	bool redflag;	bool greenflag;	bool blueflag;

	// The giant while loop below will also apply to pictures (once) as well as animations more generally: (therefore some of the above is repeated - kludgy I know (for now)).
	while ((lbitmap && rbitmap && batch==true && frame<actualFrames) || frame==0) {

		redflag=false;	greenflag=false;	blueflag=false;


		int lwidth = FreeImage_GetWidth(lbitmap);
		int lheight = FreeImage_GetHeight(lbitmap);
		int rwidth = FreeImage_GetWidth(rbitmap);
		int rheight = FreeImage_GetHeight(rbitmap);



// NEW
		
		/*if (returnDimensions==1){
			char*info = arrayc(20);
			strcpy(info, int2string(lwidth) );	strcat(info, " " );
			strcat(info, int2string(lheight) );	strcat(info, " " );
			strcat(info, int2string(rwidth) );	strcat(info, " " );
			strcat(info, int2string(rheight) );
			writefile(info,"info.txt");
			exit(0);
		}*/



		//unsigned pitch = FreeImage_GetPitch(lbitmap);
		
		int width,height;

		if (lwidth!=rwidth || lheight!=rheight) printf("Warning: Dimensions of L/R pics different! Cropping to size.\n");
		
		if (lwidth<rwidth) width=lwidth; else width=rwidth;
		if (lheight<rheight) height=lheight; else height=rheight;

		width-=(trimOut+trimIn);
		
		
		// CREATE MEMORY FOR NEW PICTURE WHICH WILL EVENTUALLY COMBINE BOTH PREVIOUS 2 PICS:
		//unsigned pitch = width*3;
		if (stereo!=true) sbitmap = FreeImage_Allocate(width, height, 24);
		else                sbitmap = FreeImage_Allocate(width*2+centerGap, height, 24);
		
		unsigned lpitch = FreeImage_GetPitch(lbitmap);
		unsigned rpitch = FreeImage_GetPitch(rbitmap);
		unsigned spitch = FreeImage_GetPitch(sbitmap);

		BYTE *lpixel = (BYTE*)FreeImage_GetBits(lbitmap);
		BYTE *rpixel = (BYTE*)FreeImage_GetBits(rbitmap);
		BYTE *spixel = (BYTE*)FreeImage_GetBits(sbitmap);





		if (hueChange!=0 || brightChange!=1 || chromaChange!=1 || reverseHues!=0) {
			// Pre-process left image to change hue or brightness (value):
			redflag=false;	greenflag=false;	blueflag=false;
			for (int y=0; y<height; y++) {
				for (int x=0; x<width; x++) {
					RGB pRGB; HSV pHSV;
					// Convert to normalized double values to allow great code simplification,
					// flexibility, and accuracy, at the slight cost of speed:
					pRGB.r = lpixel[lpitch*y + (x+trimOut)*3 + 2]/255.0;
					pRGB.g = lpixel[lpitch*y + (x+trimOut)*3 + 1]/255.0;
					pRGB.b = lpixel[lpitch*y + (x+trimOut)*3 + 0]/255.0;
					// Convert RGB representation to HSV representation:
					pHSV = toHSV(pRGB);



					// Image processing:
					if (reverseHues==1) { pHSV.h= 6.0 - pHSV.h; }
					// Image processing in the HSV colour model: (0=red, 1=yellow, 2=green, 3=cyan, 4=blue, 5=magenta)
					if (hueChange!=0) { pHSV.h +=hueChange; }
					// More image processing:
					if (brightChange!=1) { pHSV.v *=brightChange; }
					if (chromaChange!=1) { pHSV.s *=chromaChange; }
					
					

					// Convert HSV values back to RGB:
					pRGB = toRGB(pHSV);
					// Test for brighter than whites!
					if (pRGB.r>1.0) {pRGB.r=1.0; if (redflag==false)   printf("Notice: Red component truncated to 100%%\n");    redflag=true;}
					if (pRGB.g>1.0) {pRGB.g=1.0; if (greenflag==false) printf("Warning: Green component truncated to 100%%\n"); greenflag=true;}
					if (pRGB.b>1.0) {pRGB.b=1.0; if (blueflag==false)  printf("Warning: Blue component truncated to 100%%\n");  blueflag=true;}
					// Convert from float values back to integer
					lpixel[lpitch*y + (x+trimOut)*3 + 2] = (int)(pRGB.r*255.0);
					lpixel[lpitch*y + (x+trimOut)*3 + 1] = (int)(pRGB.g*255.0);
					lpixel[lpitch*y + (x+trimOut)*3 + 0] = (int)(pRGB.b*255.0);
				}
			}

			// Pre-process right image to change HSV:
			redflag=false;	greenflag=false;	blueflag=false;
			for (int y=0; y<height; y++) {
				for (int x=0; x<width; x++) {
					RGB pRGB; HSV pHSV;
					// Convert to normalized double values to allow great code simplification,
					// flexibility, and accuracy, at the slight cost of speed:
					pRGB.r = rpixel[rpitch*y + (x+trimIn)*3 + 2]/255.0;
					pRGB.g = rpixel[rpitch*y + (x+trimIn)*3 + 1]/255.0;
					pRGB.b = rpixel[rpitch*y + (x+trimIn)*3 + 0]/255.0;
					// Convert RGB representation to HSV representation:
					pHSV = toHSV(pRGB);



					// Image processing:
					if (reverseHues==1) { pHSV.h= 6.0 - pHSV.h; }
					// Image processing in the HSV colour model: (0=red, 1=yellow, 2=green, 3=cyan, 4=blue, 5=magenta)
					if (hueChange!=0) { pHSV.h +=hueChange; }
					// More image processing:
					if (brightChange!=1) { pHSV.v *=brightChange; }
					if (chromaChange!=1) { pHSV.s *=chromaChange; }



					// Convert HSV values back to RGB:
					pRGB = toRGB(pHSV);
					// Test for brighter than whites!
					if (pRGB.r>1.0) {pRGB.r=1.0; if (redflag==false)   printf("Notice: Red component truncated to 100%%\n");    redflag=true;}
					if (pRGB.g>1.0) {pRGB.g=1.0; if (greenflag==false) printf("Warning: Green component truncated to 100%%\n"); greenflag=true;}
					if (pRGB.b>1.0) {pRGB.b=1.0; if (blueflag==false)  printf("Warning: Blue component truncated to 100%%\n");  blueflag=true;}
					// Convert from float values back to integer
					rpixel[rpitch*y + (x+trimIn)*3 + 2] = (int)(pRGB.r*255.0);
					rpixel[rpitch*y + (x+trimIn)*3 + 1] = (int)(pRGB.g*255.0);
					rpixel[rpitch*y + (x+trimIn)*3 + 0] = (int)(pRGB.b*255.0);
				}
			}
		}



		if (stereo==true) {
				for (int y=0; y<height; y++) {
					
					for (int x=0; x<width; x++) {
						// Draw left picture
						spixel[spitch*y + x*3 + 2] = rpixel[rpitch*y + (x+trimIn)*3 + 2];
						spixel[spitch*y + x*3 + 1] = rpixel[rpitch*y + (x+trimIn)*3 + 1];
						spixel[spitch*y + x*3 + 0] = rpixel[rpitch*y + (x+trimIn)*3 + 0];

						// Draw right picture
						spixel[spitch*y + (x+width+centerGap)*3 + 2] = lpixel[lpitch*y + (x+trimOut)*3 + 2];
						spixel[spitch*y + (x+width+centerGap)*3 + 1] = lpixel[lpitch*y + (x+trimOut)*3 + 1];
						spixel[spitch*y + (x+width+centerGap)*3 + 0] = lpixel[lpitch*y + (x+trimOut)*3 + 0];
					}

					// Draw Center gap section
					for (int x=0; x<centerGap; x++) {
						spixel[spitch*y + (x+width)*3 + 2] = gapr;
						spixel[spitch*y + (x+width)*3 + 1] = gapg;
						spixel[spitch*y + (x+width)*3 + 0] = gapb;
					}
				}		
		}




		// Standard anaglyph process:
		if (stereo==false) {

					for (int y=0; y<height; y++) {
						for (int x=0; x<width; x++) {
							// Take left picture, put it into stereo picture and convert it to greyscale:
							int grey = (int)((lpixel[lpitch*y + (x+trimOut)*3 + 2] + lpixel[lpitch*y + (x+trimOut)*3 + 1] + lpixel[lpitch*y + (x+trimOut)*3 + 0])/3.0);
							spixel[spitch*y + x*3 + 2] = grey;
							spixel[spitch*y + x*3 + 1] = grey;
							spixel[spitch*y + x*3 + 0] = grey;

							// Optionally convert right picture to greyscale for a mono effect:
							if (mono==true) {
								grey = (int)((rpixel[rpitch*y + (x+trimIn)*3 + 2] + rpixel[rpitch*y + (x+trimIn)*3 + 1] + rpixel[rpitch*y + (x+trimIn)*3 + 0])/3.0);
								rpixel[rpitch*y + (x+trimIn)*3 + 2] = grey;
								rpixel[rpitch*y + (x+trimIn)*3 + 1] = grey;
								rpixel[rpitch*y + (x+trimIn)*3 + 0] = grey;
							}

							// Take green and blue of Right picture, and place it into stereo picture:
							spixel[spitch*y + x*3 + 1] = rpixel[rpitch*y + (x+trimIn)*3 + 1];
							spixel[spitch*y + x*3 + 0] = rpixel[rpitch*y + (x+trimIn)*3 + 0];
						}
					}

					//	sbitmap = FreeImage_ConvertToGreyscale(sbitmap);

					
					// All the usual stuff is done now. We can leave it at this, but time
					// for some post-processing to possibly enhance the 3D effect somewhat:
					
					// ************************************ POST PROCESSING
					for (int y=0; y<height; y++) {
						for (int x=0; x<width; x++) {
							// Convert to normalized double values to allow great code simplification,
							// flexibility, and accuracy, at the slight cost of speed:
							double r,g,b;
							double pr = spixel[spitch*y + x*3 + 2]/255.0;
							double pg = spixel[spitch*y + x*3 + 1]/255.0;
							double pb = spixel[spitch*y + x*3 + 0]/255.0;
							r=pr; g=pg; b=pb;

							/*
							if (cyanChange!=1) {
								// Change Cyan component:
								g=pow(g,cyanChange);
								b=pow(b,cyanChange);
								// Reduce green and blue level for above Cyan adjustment:
								//g +=0.5*(1 -cyan); b +=0.5*(1 -cyan);  // old routine, but makes too greyish
								g *= pow(cyanChange,0.5);  // 
								b *= pow(cyanChange,0.5);
							}
							*/

							if (redChange!=1) {
								// Increase red component optionally: best = ~1.4
								r*=redChange;
							}
							

							// Experimental adjustment to reduce anaglyph 'ghosting' (filter pollution)
							if (noEnhance==0) {
								r*=pow(1.0-(g/evar1),evar2); // 1.2 0.2
								g*=pow(1.0-(g/evar1),evar2); // 1.2 0.2
							}

							// Test for brighter than whites!
							if (r>1.0) {r=1.0; if (redflag==false)   printf("Notice: Red component truncated to 100%%\n");    redflag=true;}
							if (g>1.0) {g=1.0; if (greenflag==false) printf("Warning: Green component truncated to 100%%\n"); greenflag=true;}
							if (b>1.0) {b=1.0; if (blueflag==false)  printf("Warning: Blue component truncated to 100%%\n");  blueflag=true;}

							// Convert from float values back to integer
							spixel[spitch*y + x*3 + 2] = (int)(r*255.0);
							spixel[spitch*y + x*3 + 1] = (int)(g*255.0);
							spixel[spitch*y + x*3 + 0] = (int)(b*255.0);
						}
					}
					// End of POST-PROCESSING section
					// ********************************************************


		}

		double ratio = width/388.0;	double smaller=3.0;		int wmx = (int)(width/smaller); int wmy = (int)(77/smaller*ratio); 

		//wmsbitmap = FreeImage_Allocate(wmx, wmy, 24);

		wmsbitmap = FreeImage_Rescale(wmbitmap, (int)(width/smaller), (int)(77/smaller*ratio), FILTER_BILINEAR);
		FreeImage_Paste(sbitmap, wmsbitmap, (int)(width/2.0-388/2.0/smaller*ratio), (int)(height-(90.0/smaller*ratio)), 128);



		char outputStereo[300];

		strcpy(outputStereo, "Preview/stereo_out");
		if (batch==true) 	strcat(outputStereo,getFileNumber(frame,left));
		strcat(outputStereo,".");
		strcat(outputStereo,saveFormat);

		// Saving of file:
		int save=0;
		if (strcmp(saveFormat,"gif")==0 || col256==1) {
			// NeuQuant neural-net quantization algorithm by Anthony Dekker
			sbitmap = FreeImage_ColorQuantize(sbitmap, FIQ_NNQUANT); // FIQ_WUQUANT is faster but less accurate.
			//sbitmap = FreeImage_ConvertTo8Bits(sbitmap);
		}
		if (strcmp(saveFormat,"png")==0) save=FreeImage_Save(FIF_PNG, sbitmap, outputStereo, 0);
		if (strcmp(saveFormat,"bmp")==0) save=FreeImage_Save(FIF_BMP, sbitmap, outputStereo, 0);
		if (strcmp(saveFormat,"jpg")==0) save=FreeImage_Save(FIF_JPEG, sbitmap, outputStereo, 0);

		if (strcmp(saveFormat,"gif")==0) {
			save=FreeImage_Save(FIF_GIF, sbitmap, outputStereo, 0);
		}

		if (save) {
			printf("Saving frame: %i\n",frame);
		}
		else {
			printf("PNG, BMP, JPG, or GIF formats allowed only.");
			return 0;
		}

		

		frame++;

		if (batch==true) {
			
			FreeImage_Unload(lbitmap);
			FreeImage_Unload(rbitmap);
			FreeImage_Unload(sbitmap);
			FreeImage_Unload(wmsbitmap);

			if (swapxy==false) {
				if ( stricmp(lpicformat,"png")==0 ) {	lbitmap = FreeImage_Load(FIF_PNG, incrementFile(frame, left), PNG_DEFAULT);	rbitmap = FreeImage_Load(FIF_PNG, incrementFile(frame, right), PNG_DEFAULT);  }
				if ( stricmp(lpicformat,"bmp")==0 ) {	lbitmap = FreeImage_Load(FIF_BMP, incrementFile(frame, left), BMP_DEFAULT);	rbitmap = FreeImage_Load(FIF_BMP, incrementFile(frame, right), BMP_DEFAULT);  }
				if ( stricmp(lpicformat,"gif")==0 ) {	lbitmap = FreeImage_Load(FIF_GIF, incrementFile(frame, left), GIF_DEFAULT);	rbitmap = FreeImage_Load(FIF_GIF, incrementFile(frame, right), GIF_DEFAULT);  }
				if ( stricmp(lpicformat,"peg")==0 || stricmp(lpicformat,"jpg")==0) {	lbitmap = FreeImage_Load(FIF_JPEG, left, JPEG_DEFAULT);	rbitmap = FreeImage_Load(FIF_JPEG, right, JPEG_DEFAULT);  }
			} else {
				if ( stricmp(lpicformat,"png")==0 ) {	lbitmap = FreeImage_Load(FIF_PNG, incrementFile(frame, right), PNG_DEFAULT);	rbitmap = FreeImage_Load(FIF_PNG, incrementFile(frame, left), PNG_DEFAULT);  }
				if ( stricmp(lpicformat,"bmp")==0 ) {	lbitmap = FreeImage_Load(FIF_BMP, incrementFile(frame, right), BMP_DEFAULT);	rbitmap = FreeImage_Load(FIF_BMP, incrementFile(frame, left), BMP_DEFAULT);  }
				if ( stricmp(lpicformat,"gif")==0 ) {	lbitmap = FreeImage_Load(FIF_GIF, incrementFile(frame, right), GIF_DEFAULT);	rbitmap = FreeImage_Load(FIF_GIF, incrementFile(frame, left), GIF_DEFAULT);  }
				if ( stricmp(lpicformat,"jpg")==0 || stricmp(lpicformat,"peg")==0) {	lbitmap = FreeImage_Load(FIF_JPEG, incrementFile(frame, right), JPEG_DEFAULT);	rbitmap = FreeImage_Load(FIF_JPEG, incrementFile(frame, left), JPEG_DEFAULT);  }
			}
	
			if (lbitmap && rbitmap) {
				printf(" %i bit color. ",FreeImage_GetBPP(lbitmap));
				if (FreeImage_GetBPP(lbitmap)!=24 || FreeImage_GetBPP(rbitmap)!=24) {
					printf(" Converting to 24 bit. ");
					lbitmap = FreeImage_ConvertTo24Bits(lbitmap);
					rbitmap = FreeImage_ConvertTo24Bits(rbitmap);
				}
			}
		}


	}

	if (batch!=true) {
		FreeImage_Unload(lbitmap);
		FreeImage_Unload(rbitmap);
		FreeImage_Unload(sbitmap);
		FreeImage_Unload(wmbitmap);
		FreeImage_Unload(wmsbitmap);
	}

	printf("\nSuccess!\n\n");
	printf("There is a watermark in each image. Please make a donation of your choice ($5 to $20 recommended\n");
	printf("if you find want it removed. See http://www.skytopia.com/software/stereoptica for details.\n");
	
	
	
	return 0;
}






// Super useful generic funntion to increment a filename with a number in it.
// E.g. filename001.png could go to filename002.png or my9file.txt could go to
// my10file.txt.
// int i is the amount to increment, and char*d is the file path/name.
char* incrementFile(int i, char*d) {
	char*newFile = (char*)malloc(300*sizeof(char));

	int len=strlen(d);

	int nend,nbeg;
	bool flag=false;
	bool succeed=false;
	int n;
	for (n=len-1; n>=-1; n--) {

		if (d[n]-48>=0 && d[n]-48<10 && flag==false) {
			nend=n;
			flag=true;			
		}
		if (flag==true) {
			if (  !(d[n]-48>=0 && d[n]-48<10)   ) {
				nbeg=n;
				succeed=true;
				break;
			}
		}
		//printf("a: %c ",d[n]);
		
	}

	//if (n==-2 && succeed==false) {printf("Bad filename"); return "Bad filename";}
	if (n==-2 && succeed==false) {
		printf("Bad filename - No number in filename for batch processing");

		writefile("-1","info.txt");

		exit(0);
		
		//return "Bad filename";
	}	

	char number[15];
	int m;
	for (m=1; m<=nend-nbeg; m++ ) {		
		number[m-1]=d[nbeg+m];
	}
	number[m-1]='\0';

	int figures=strlen(number);
	int newNumber = atoi(number);

	newNumber+=i;

	for (m=0; m<=nbeg; m++ ) 	{newFile[m] = d[m];   }
	
	if (figures==1) sprintf(number, "%01d", newNumber);
	if (figures==2) sprintf(number, "%02d", newNumber);
	if (figures==3) sprintf(number, "%03d", newNumber);
	if (figures==4) sprintf(number, "%04d", newNumber);
	if (figures==5) sprintf(number, "%05d", newNumber);
	if (figures==6) sprintf(number, "%06d", newNumber);
	if (figures==7) sprintf(number, "%07d", newNumber);
	if (figures==8) sprintf(number, "%08d", newNumber);

	figures=strlen(number);

	for (m=1; m<=figures; m++ ) {newFile[m+nbeg]=number[m-1]; }
	
	int c=nbeg+figures+1;
	for (m=nend+1; m<=len; m++ ) { newFile[c]=d[m];   c++; 	}

	newFile[c]='\0';	
	return newFile;
}




// Gets number from a filename:
char* getFileNumber(int i, char*d) {
	char*newFile = (char*)malloc(300*sizeof(char));
	char*number = (char*)malloc(15*sizeof(char));

	int len=strlen(d);

	int nend,nbeg;
	bool flag=false;
	bool succeed=false;
	int n;
	for (n=len-1; n>=-1; n--) {

		if (d[n]-48>=0 && d[n]-48<10 && flag==false) {
			nend=n;
			flag=true;			
		}
		if (flag==true) {
			if (  !(d[n]-48>=0 && d[n]-48<10)   ) {
				nbeg=n;
				succeed=true;
				break;
			}
		}
		//printf("a: %c ",d[n]);
		
	}
	if (n==-2 && succeed==false) {printf("Bad filename"); return "Bad filename";}
	
	int m;
	for (m=1; m<=nend-nbeg; m++ ) {		
		number[m-1]=d[nbeg+m];
	}
	number[m-1]='\0';

	int figures=strlen(number);
	int newNumber = atoi(number);

	newNumber+=i;

	for (m=0; m<=nbeg; m++ ) 	{newFile[m] = d[m];   }
	
	if (figures==1) sprintf(number, "%01d", newNumber);
	if (figures==2) sprintf(number, "%02d", newNumber);
	if (figures==3) sprintf(number, "%03d", newNumber);
	if (figures==4) sprintf(number, "%04d", newNumber);
	if (figures==5) sprintf(number, "%05d", newNumber);
	if (figures==6) sprintf(number, "%06d", newNumber);
	if (figures==7) sprintf(number, "%07d", newNumber);
	if (figures==8) sprintf(number, "%08d", newNumber);
	
	return number;
}


RGB toRGB(HSV z) {
	RGB n;

	if (z.s>1) {z.s=1; }

	z.h = mod(z.h,6.0);

	int h2 = (int)(z.h) ;
	double f = z.h - h2;
	double p = z.v*(1-z.s);
	double q = z.v*(1-f*z.s);
	double t = z.v*(1-(1-f)*z.s);

	if (h2==0) { n.r=z.v; n.g=t; n.b=p;}
	if (h2==1) { n.r=q; n.g=z.v; n.b=p;}
	if (h2==2) { n.r=p; n.g=z.v; n.b=t;}
	if (h2==3) { n.r=p; n.g=q; n.b=z.v;}
	if (h2==4) { n.r=t; n.g=p; n.b=z.v;}
	if (h2==5) { n.r=z.v; n.g=p; n.b=q;}

	return n;
}


HSV toHSV(RGB z) {
	HSV n;

// a < b ? a : b
	double min = ((z.r <? z.g) <? z.b);
	double max = ((z.r >? z.g) >? z.b);


	if (max==min) n.h = 0;
	else {
		if (max==z.r && z.g>=z.b) n.h = (z.g-z.b)/(max-min)+0;
		if (max==z.r && z.g<z.b) n.h = (z.g-z.b)/(max-min)+6;
		if (max==z.g) n.h = (z.b-z.r)/(max-min)+2;
		if (max==z.b) n.h = (z.r-z.g)/(max-min)+4;
	}

	if (max==0) n.s=0;
	else n.s = (max-min)/max;
	
	n.v=max;
//	if (n.s>1) {n.s=1; }

	return n;
}

double mod(double a, double b) {
	double m = a-b*((int)(a/b));
	if (m<0) m+=b;
	return m;
}
